;
; Highest nibble is not decoded and not significant
;
sd_port	equ	$F80000

boot_to	equ	$8000
;SD_CMD0	equ	$02	; Go idle
SD_CMD0	equ	$40	; Go idle
SD_CMD1	equ	$41	; SEND_OP_COND (MMC) Activates card's init process
SD_CMD8	equ	$48	; 
SD_CMD10	equ	$4a	; 
SD_CMD16	equ	$50	; Set block length
SD_CMD17	equ	$51	; Read sector

BLOCKSIZE_H	equ	$02	; Means SD block size is 512
BLOCKSIZE_L	equ	$00	;
BLOCKSIZE	equ 	$200

f_boot:
  movea.l	#f_boot_msg,a0	; saying "we are booting"
  bsr	tx_str
  bsr	sd_cs_high	; sending 80 ticks when CS high
  move.b		#$80,d1
  bsr	sd_send_ticks_cs_high
  bsr	sd_cs_low
; sending CMD0
  move.b		#SD_CMD0,d1
  bsr	sd_write_cmd
  movea.l	#f_boot_cmd0_done_msg,a0	; telling about that
  bsr	tx_str

; waiting for 01 response for $100 times
  move.w		#$100,d4
f_boot_cmd0_loop:
  bsr	sd_read_byte
  cmpi.b		#$01,d0	; waiting for 01 
  beq	f_boot_cmd0_got_resp	; ok, got it
  dbra	d4,f_boot_cmd0_loop	; retrying
  movea.l	#f_boot_cmd0_no_resp_msg,a0	; oh no - failed
  bsr	tx_str	; telling that ang exiting
  rts
f_boot_cmd0_got_resp:
  movea.l	#f_boot_cmd0_resp_msg,a0	; saying we've got CMD0 response "Cmd0 got 01"
  bsr	tx_str

; so sending CMD1s till we get 00 (not 01!)
  move.w		#$100,d4
f_boot_cmd1_loop:
  move.w		d4,-(sp)
  move.b	#SD_CMD1,d1	; so sending CMD1
  bsr	sd_write_cmd_and_get_result
  move.w	(sp)+,d4
  cmpi.b		#$00,d0	; waiting for 00 
  beq	f_boot_cmd1_got_resp	; ok, got it
  dbra	d4,f_boot_cmd1_loop	; retrying
  movea.l	#f_boot_cmd1_no_resp_msg,a0	; oh no - failed
  bsr	tx_str	; telling that ang exiting
  rts
f_boot_cmd1_got_resp:
  movea.l	#f_boot_cmd1_resp_msg,a0	; saying we've got CMD1 response
  bsr	tx_str

  move.b	#0,d2	; sending CMD16 - "set block size to 512 bytes"
  move.b	#0,d3	; that is not absoulutely needed, as I understand
  move.b	#BLOCKSIZE_H,d4
  move.b	#BLOCKSIZE_L,d5
  move.b	#SD_CMD16,d1
  bsr	sd_write_cmd_with_args_and_get_result
  cmpi.b		#$00,d0	;  sector should be 512 
  beq	f_boot_cmd16_ok
  movea.l	#f_boot_cmd16_not_ok_msg,a0
  bsr	tx_str	
  rts
f_boot_cmd16_ok:
  movea.l	#f_boot_cmd16_ok_msg,a0
  bsr	tx_str	

; CMD 10

  move.b	#SD_CMD10,d1 ; fetching card's name
  bsr	sd_write_cmd_and_get_result
  move.w	#$10,d3
f_boot_cmd10_loop:
  move.w	d3,-(sp)
  bsr	sd_read_byte
  move.w	(sp)+,d3
  cmpi.b	#$FE,d0	; waiting for FE
  beq		f_boot_cmd10_done
  dbra	d3,f_boot_cmd10_loop

  movea.l	#f_boot_cmd10_not_ok_msg,a0
  bsr	tx_str	
  rts
f_boot_cmd10_done:
  movea.l	#f_boot_cmd10_got_fe,a0
  bsr	tx_str	


  bsr	sd_read_byte	; skipping manufacturer ID+smth else
  bsr	sd_read_byte
  bsr	sd_read_byte

  movea.l	#f_boot_card_id_msg,a0 ; saying "Card ID:"
  bsr	tx_str	
  move.w	#$4,d3	; next 5 bytes are ASCII Card name
f_boot_cmd10_done_loop2:
  move.w	d3,-(sp)
  bsr	sd_read_byte
  bsr	tx_char
  move.w	(sp)+,d3
  dbra	d3,f_boot_cmd10_done_loop2	; 
  move.b	#CR,d0
  bsr	tx_char
  move.b	#LF,d0
  bsr	tx_char
  move.w	#$B,d3	; skip next 11 bytes
f_boot_cmd10_done_loop3:
  move.w	d3,-(sp)
  bsr	sd_read_byte
  move.w	(sp)+,d3
  dbra	d3,f_boot_cmd10_done_loop3	; 
; actually booting sector 0

  movea.l	#boot_to,a5	; where we will boot it

  move.w	#$1F,d6	; will repeat 32 times = 16K
  move.b	#$0,d4  ; d2 - always 0, d3 - always 0, d4 - from 0 to 40, d5 - 0
f_boot_sector_loop:

  move.b	d6,-(sp)
  move.b	d4,-(sp)
  move.b	#$0,d2
  move.b	#$0,d3
  move.b	#$0,d5
  move.b	#SD_CMD17,d1 ; Reading actually
  bsr	sd_write_cmd_with_args_and_get_result

  move.w	#$100,d3	; waiting for data token
f_boot_cmd17_loop:
  move.w	d3,-(sp)
  bsr	sd_read_byte
  move.w	(sp)+,d3
  cmpi.b	#$FE,d0	; waiting for FE
  beq		f_boot_cmd17_done
  dbra	d3,f_boot_cmd17_loop

  move.b	(sp)+,d4 ; cannot get FE - popping D4
  move.b	(sp)+,d4 ; cannot get FE - popping D6
  movea.l	#f_boot_cmd17_no_fe,a0
  bsr	tx_str
  rts

f_boot_cmd17_done:	; read data token done
  move.w	#BLOCKSIZE-1,d4 ; 512 bytes
f_boot_sector_read:
  move.w	d4,-(sp)
  bsr	sd_read_byte
  move.b	d0,(a5)+
;  move.b	#'*',d0
;  bsr	tx_char
  move.w	(sp)+,d4
  dbra	d4,f_boot_sector_read

  bsr	sd_read_byte	; reading and ignoring checksumm
  bsr	sd_read_byte	; again, checking it is a good idea

  move.l	#f_boot_sector_read_msg,a0
  bsr	tx_str
 
  move.b	(sp)+,d4
  addi.b	#2,d4
  move.b	(sp)+,d6
  dbra	d6,f_boot_sector_loop

  rts


 


; send d1 ticks to sd card when cs high. Used by init.
sd_send_ticks_cs_high:
  move.b		d1,d7
  move.w		#7,sd_port	; cs=1 data=1 clock=up
  move.b		#'^',d0
  bsr	tx_char
  move.b		#6,sd_port	; cs=1 data=1 clock=low
  move.b		#'v',d0
  bsr	tx_char
  move.b	d7,d1
  dbra	d1,sd_send_ticks_cs_high
  rts
  

; writes command in D1 ; returns D0 - ff if error, result if not
sd_write_cmd_and_get_result:
  bsr	sd_write_cmd
sd_write_cmd_and_get_result_preloop:
  move.w		#$40,d1	; 64 tries in worst case
sd_write_cmd_and_get_result_loop:  
  move.w		d1,-(sp)
  bsr	sd_read_byte	; returns in D0
  move.w		(sp)+,d1
  cmpi.b		#$ff,d0
  bne	sd_write_cmd_and_get_result_done
  dbra	d1,sd_write_cmd_and_get_result_loop
sd_write_cmd_and_get_result_done:
  rts


; writes SD command D1, with no arguments
sd_write_cmd:
; should be there write FF? That's be good
; IT MUST!
  move.w		d1,-(sp)
  move.b	#$ff,d1
  bsr	sd_write_byte
  move.w		(sp)+,d1
; or may be ^^ that's not needed.
  bsr	sd_write_byte
  clr.b	d1
  bsr	sd_write_byte
  clr.b	d1
  bsr	sd_write_byte
  clr.b	d1
  bsr	sd_write_byte
  clr.b	d1
  bsr	sd_write_byte
  move.b		#$95,d1
  bsr	sd_write_byte
  rts

; writes command in D1 ; returns D0 - ff if error, result if not
; arguments d2,d3,d4,d5
sd_write_cmd_with_args_and_get_result:
  bsr	sd_write_cmd_with_args
  bra	sd_write_cmd_and_get_result_preloop ; the code is equal


; writes SD command D1, with  arguments d2,d3,d4,d5
sd_write_cmd_with_args:
; should be there "write FF?" That's be good
  move.b	d5,-(sp)
  move.b	d4,-(sp)
  move.b	d3,-(sp)
  move.b	d2,-(sp)
  move.b	d1,-(sp)
  move.b	#$ff,d1
  bsr	sd_write_byte
  move.b		(sp)+,d1
; or may be ^^ that's not needed.
  bsr	sd_write_byte
  move.b		(sp)+,d1
  bsr	sd_write_byte
  move.b		(sp)+,d1
  bsr	sd_write_byte
  move.b		(sp)+,d1
  bsr	sd_write_byte
  move.b		(sp)+,d1
  bsr	sd_write_byte
  move.b		#$95,d1
  bsr	sd_write_byte
  rts


; writes byte D1
sd_write_byte:
;  andi.l	#$000000FF,d1 ; that should be byte ; but looks like that's not needed

; delete this debug code
  move.b		d1,d7
  move.b		#'[',d0
  bsr	tx_char
  move.b		#'W',d0
  bsr	tx_char
  move.b		#':',d0
  bsr	tx_char
  move.b		d7,d0
  bsr	tx_b
  move.b		d7,d1
; down to here
  move.l	d3,-(sp)
  move.w		#7,d3	; 8 bits in byte
sd_write_bit_loop:
  move.b		#0,d0
  add.b		d1,d1
  addx.b		d0,d0	; should add X==Carry flag,
  add.b		d0,d0	; 2 or 0
  move.b		d0,sd_port ; cs=0 data=data clock=0
  add.b		#$1,d0
  move.b		d0,sd_port ; cs=0 data=data clock=1
  subi.b	#$1,d0
  move.b		d0,sd_port ; cs=0 data=data clock=0

; this debug code should be deleted
;  move.b		d0,d7	; 
;  move.b		d1,d6	; 
;  move.b		#'.',d0
;  bsr	tx_char
;  move.b		d7,d0
;  addi.b		#'0',d7
;  bsr	tx_char  
;  move.b		d6,d1
;  move.b		d7,d0 ; restoring d0 d1
; -- up to here; no need to restore d0 bte

  dbra	d3,sd_write_bit_loop
  move.l	(sp)+,a3
  move.b		#']',d0
  bsr	tx_char
  rts

; returns on D0
sd_read_byte:
  move.w		#7,d3

  move.b		#$0,d0	; here will be byte stored
sd_read_byte_read_bit:
  add.b		d0,d0	; shifting it left
  move.b		#3,sd_port	; cs=0,data=1,clk=1

  move.b		sd_port,d1	; reading inverted data
  andi.b		#1,d1	; the lower bit is input
  eori.b		#1,d1	; we have an inverted data

  add.b		d1,d0	; adding to result

;  move.b		d0,d7
;  addi.b		#'0',d1  
;  move.b		d1,d0
;  bsr	tx_char
;  move.b		d7,d0

  move.b		#2,sd_port	; cs=0,data=1,clk=0


  dbra	d3,sd_read_byte_read_bit

;  move.b		d0,d7
;  move.b		#'(',d0
;  bsr	tx_char
;  move.b		#'R',d0
;  bsr	tx_char
;  move.b		#':',d0
;  bsr	tx_char
;  move.b		d7,d0
;  bsr	tx_b
;  move.b		#')',d0
;  bsr	tx_char
;  move.b		d7,d0
  rts

; Uppers CS
sd_cs_high:
  move.b		#6,sd_port	; cs=1,data=1,clk=0
  rts
; Lowers CS
sd_cs_low:
  move.b		#2,sd_port	; cs=0,data=1,clk=0
  rts

f_boot_msg:
  dc.b	'Booting',CR,LF,0
f_boot_cmd0_done_msg:
  dc.b	'CMD0 send',CR,LF,0
f_boot_cmd0_no_resp_msg:
  dc.b	'CMD0 no response',CR,LF,0
f_boot_cmd0_resp_msg:
  dc.b	'CMD0 got 01!',CR,LF,0
f_boot_cmd1_no_resp_msg:
  dc.b	'CMD1 no response',CR,LF,0
f_boot_cmd1_resp_msg:
  dc.b	'CMD1 got 00',CR,LF,0
f_boot_cmd16_ok_msg:
  dc.b	'CMD16 got 00',CR,LF,0
f_boot_cmd16_not_ok_msg:
  dc.b	'CMD16 failed - cannot set 512 b/sec',CR,LF,0
f_boot_cmd10_not_ok_msg:
  dc.b	'CMD10 failed - cannot get FE',CR,LF,0
f_boot_cmd10_got_fe:
  dc.b	'CMD10 got FE',CR,LF,0
f_boot_card_id_msg:
  dc.b	'Card ID:',0
f_boot_sector_read_msg:
  dc.b	'Sector Read',CR,LF,0
f_boot_cmd17_no_fe:
  dc.b	'Cannot get data token for CMD17',CR,LF,0
  even