/* Takes next byte of .asm program and feeds it to CPU */

#include <inttypes.h>
#include <avr/pgmspace.h>
#include "main.h"
//#include <stdio.h>
// This is the first bytes send to 68k CPU,
// generally (word) starting address, (word) stack pointer, which generally
// located at $0 and $4, and some other pre-loop commands (prologue)

uint8_t const PROGMEM prologue[] = {
  0x00,0x00,0x00,0x08                          // 0       dc.l start
  ,0x00,0x00,0xF0,0x00                          // 0       dc.l smth (stack - F000 ; why not 1FFFE ?)
                                    // 0  start:
  ,0x20,0x7C ,0x00,0x00,0x00,0x00                     // 0      movea.l<--->#0,A0
};

// this is main loop. note that -1 is substituted with actual byte of loading program
// note, it is longer then byte, int_16?

// Do not forget it must be ended by NOP, because CPU prefetches command, and if you 
// send RESET right after the last command it will not be executed
uint8_t const PROGMEM program_loop_bytes[] = {
 0x30, 0xFC, 0, 0 // move.w #XX,(A0)+
//, 0x4e, 0x71 // NOP (gone to epilogue)
}; 

uint8_t const PROGMEM epilogue[] = { // That is needed because of prefetching.
 0x4e, 0x71 // NOP
};

// The actual m68k program which is loaded into memory
uint8_t const PROGMEM m68program[] = {
#include "68prog.c"
};

// Returns -1 when all done (which should cause m68k reset), or byte to feed 68k cpu
//
int bytefeeder()
{
#ifdef DEBUG
 static unsigned const  PROGMEM char statemsg[] = "\n\rState ";
 static unsigned const  PROGMEM char nr[] = "\n\r";
// static unsigned PROGMEM char byte[] = "Byte ";
 static unsigned const  PROGMEM char progbyte[] = "\n\rByte from program";
 static unsigned const  PROGMEM char normalbyte[] = "\n\rRegular byte";
 static unsigned const  PROGMEM char progpend[] = "\n\rProgP ended";
 static unsigned const  PROGMEM char defaultmsg[] = "\n\rDefault point";
#endif

// Does not work as should
#ifdef ROTATING_SHIT
 static const PROGMEM char rotating_shit[] = "\\|/-";
 static char rotating_shit_pointer = -1;

//  static unsigned const PROGMEM char rotating_shit_vt100_1[] = "\x1B[1A";
#endif

 static unsigned char state = 0;
 static int chainp = 0, programp = 0;
 int r = 0;
#ifdef DEBUG
// USART_PUTS_my(statemsg);
// uart_puthex_byte(state);
// USART_PUTS_my(nr);
#endif

#ifdef ROTATING_SHIT

// USART_PUTCH(27);
// USART_PUTS_my(rotating_shit_vt100_1);
 USART_PUTCH(pgm_read_byte(rotating_shit+( rotating_shit_pointer++ & 3 )));
 USART_PUTCH('\r');
// USART_PUTCH('\n');
#else
// USART_PUTCH('.');
#endif
 

 if (state == 0) 
 {
 // we are loading prologue
  if (chainp < sizeof(prologue)/sizeof(prologue[0])) 
  {
   return((int) pgm_read_byte( prologue + chainp++ ));
  }
  state = 1; // done loading prologue
  chainp = 0;
  programp = 0; // and start loading loop right now (without break)
 }
 if (state == 1)
 {
  if (chainp >= (sizeof(program_loop_bytes)/sizeof(program_loop_bytes[0])))
  { // is that the last byte of loop?
   if(programp >= sizeof(m68program)/sizeof(m68program[0])) {
#ifdef DEBUG
//    USART_PUTS_my(progpend);
#endif
#ifndef ROTATING_SHIT
    USART_PUTCH('\r');
    USART_PUTCH('\n');
#endif
//    return(-1); // not only the last byte of loop, but the end of 68k program!
    state = 2;
   }
   chainp = 0; // lets repeat the loop
  }
  if (state != 2) { // if we got to the end of program to epilogue, skip this
//   r = pgm_read_word( program_loop_bytes + chainp++);
   r = pgm_read_byte( program_loop_bytes + chainp++);
   if (r == 0) //should we sustitute the byte with the byte of actual program?
   {
#ifdef DEBUG
//   USART_PUTS_my(progbyte);
#endif
    return((int)(pgm_read_byte(m68program+programp++)));
   }
#ifdef DEBUG
//  USART_PUTS_my(normalbyte);
#endif
   return(r);
  }
 }
 if (state == 2) 
 {
 // we are loading epilogue. That is needed because of prefetching.
  if (chainp < sizeof(epilogue)/sizeof(epilogue[0])) 
  {
   return((int) pgm_read_byte( epilogue + chainp++ ));
  }
  return(-1); // All done
 }

// This should never happen
#ifdef DEBUG
 USART_PUTS_my(defaultmsg);
#endif
 return(r);
}

/*
int main()
{
 int b;
 while ((b=bytefeeder()) != -1) 
 {
  printf("[%02X]\n",b);
 }
 return(0);
}

*/
